@@ -29,12 +29,17 @@
 #include <QDBusArgument>
 #include <QDesktopServices>
 #include <QLabel>
+#include <QNetworkCookie>
+#include <QNetworkReply>
 #include <QNetworkRequest>
+#include <QPointer>
 #include <QProgressBar>
 #include <QPushButton>
 #include <QRegExp>
 #include <QSettings>
+#include <QSslError>
 #include <QStackedLayout>
+#include <QTimer>
 #include <QVBoxLayout>
 #include <QWebElement>
 #include <QWebFrame>
@@ -46,6 +51,8 @@ using namespace SignOnUi;
 namespace SignOnUi {
 
 static const QString keyPreferredWidth = QString("PreferredWidth");
+static const QString keyHorizontalScrollBar = QString("HorizontalScrollBar");
+static const QString keyVerticalScrollBar = QString("VerticalScrollBar");
 static const QString keyTextSizeMultiplier = QString("TextSizeMultiplier");
 static const QString keyUserAgent = QString("UserAgent");
 static const QString keyViewportWidth = QString("ViewportWidth");
@@ -56,10 +63,14 @@ static const QString keyPasswordField = QString("PasswordField");
 static const QString keyLoginButton = QString("LoginButton");
 static const QString keyInternalLinksPattern = QString("InternalLinksPattern");
 static const QString keyExternalLinksPattern = QString("ExternalLinksPattern");
+static const QString keyAllowedUrls = QString("AllowedUrls");
+static const QString valueAlwaysOn = QString("alwaysOn");
+static const QString valueAlwaysOff = QString("alwaysOff");
 
 /* Additional session-data keys we support. */
 static const QString keyCookies = QString("Cookies");
 static const QString keyAllowedSchemes = QString("AllowedSchemes");
+static const QString keyIgnoreSslErrors = QString("IgnoreSslErrors");
 
 class WebPage: public QWebPage
 {
@@ -85,6 +96,11 @@ public:
         m_allowedSchemes = schemes;
     }
 
+    void setAllowedUrls(const QString &pattern) {
+        m_allowedUrls =
+            QRegExp(pattern, Qt::CaseInsensitive, QRegExp::RegExp2);
+    }
+
     void setFinalUrl(const QUrl &url) { m_finalUrl = url; }
 
 protected:
@@ -135,15 +151,24 @@ private:
     QRegExp m_externalLinksPattern;
     QRegExp m_internalLinksPattern;
     QStringList m_allowedSchemes;
+    QRegExp m_allowedUrls;
     QUrl m_finalUrl;
 };
 
 bool WebPage::urlIsBlocked(QUrl url) const {
+    if (url == QUrl("about:blank")) return false;
+
     if (!m_allowedSchemes.contains(url.scheme())) {
         TRACE() << "Scheme not allowed:" << url.scheme();
         return true;
     }
 
+    if (!m_allowedUrls.isEmpty() &&
+        !m_allowedUrls.exactMatch(url.toString())) {
+        TRACE() << "URL not allowed:" << url;
+        return true;
+    }
+
     QString urlText = url.toString(QUrl::RemoveScheme |
                                    QUrl::RemoveUserInfo |
                                    QUrl::RemoveFragment |
@@ -217,8 +242,11 @@ public:
     void start();
 
 private Q_SLOTS:
+    void onSslErrors(QNetworkReply *reply, const QList<QSslError> &errors);
     void onUrlChanged(const QUrl &url);
+    void onLoadProgress();
     void onLoadFinished(bool ok);
+    void onFailTimer();
     void onFinished();
     void startProgress();
     void stopProgress();
@@ -237,7 +265,10 @@ private:
 
 private:
     mutable BrowserRequest *q_ptr;
-    Dialog *m_dialog;
+    /* The dialog can be deleted by the Request class, if it's set as children
+     * of an embedded widget which is then deleted. Therefore, in order to
+     * avoid a double deletion, guard the pointer with a QPointer. */
+    QPointer<Dialog> m_dialog;
     QStackedLayout *m_dialogLayout;
     QWidget *m_webViewPage;
     QWidget *m_successPage;
@@ -255,6 +286,8 @@ private:
     QString m_username;
     QString m_password;
     int m_loginCount;
+    bool m_ignoreSslErrors;
+    QTimer m_failTimer;
 };
 
 } // namespace
@@ -267,8 +300,13 @@ BrowserRequestPrivate::BrowserRequestPrivate(BrowserRequest *request):
     m_webView(0),
     m_animationLabel(0),
     m_settings(0),
-    m_loginCount(0)
+    m_loginCount(0),
+    m_ignoreSslErrors(false)
 {
+    m_failTimer.setSingleShot(true);
+    m_failTimer.setInterval(3000);
+    QObject::connect(&m_failTimer, SIGNAL(timeout()),
+                     this, SLOT(onFailTimer()));
 }
 
 BrowserRequestPrivate::~BrowserRequestPrivate()
@@ -276,11 +314,21 @@ BrowserRequestPrivate::~BrowserRequestPrivate()
     delete m_dialog;
 }
 
+void BrowserRequestPrivate::onSslErrors(QNetworkReply *reply,
+                                        const QList<QSslError> &errors)
+{
+    TRACE() << errors;
+    if (m_ignoreSslErrors) {
+        reply->ignoreSslErrors();
+    }
+}
+
 void BrowserRequestPrivate::onUrlChanged(const QUrl &url)
 {
     Q_Q(BrowserRequest);
 
     TRACE() << "Url changed:" << url;
+    m_failTimer.stop();
 
     if (url.host() == finalUrl.host() &&
         url.path() == finalUrl.path()) {
@@ -297,12 +345,17 @@ void BrowserRequestPrivate::onUrlChanged(const QUrl &url)
     setupViewForUrl(url);
 }
 
+void BrowserRequestPrivate::onLoadProgress()
+{
+    m_failTimer.stop();
+}
+
 void BrowserRequestPrivate::onLoadFinished(bool ok)
 {
     TRACE() << "Load finished" << ok;
 
     if (!ok) {
-        notifyLoadFailed();
+        m_failTimer.start();
         return;
     }
 
@@ -323,6 +376,11 @@ void BrowserRequestPrivate::onLoadFinished(bool ok)
     }
 }
 
+void BrowserRequestPrivate::onFailTimer()
+{
+    notifyLoadFailed();
+}
+
 void BrowserRequestPrivate::addBrowserCookies(CookieJar *cookieJar)
 {
     Q_Q(BrowserRequest);
@@ -390,6 +448,9 @@ QWidget *BrowserRequestPrivate::buildWebViewPage(const QVariantMap &params)
     WebPage *page = new WebPage(this);
     QObject::connect(page, SIGNAL(contentsChanged()),
                      this, SLOT(onContentsChanged()));
+    QObject::connect(page->networkAccessManager(),
+                     SIGNAL(sslErrors(QNetworkReply*,const QList<QSslError> &)),
+                     this, SLOT(onSslErrors(QNetworkReply*,const QList<QSslError> &)));
     m_webView->setPage(page);
 
     /* The following couple of lines serve to instruct the QWebPage not to load
@@ -401,10 +462,7 @@ QWidget *BrowserRequestPrivate::buildWebViewPage(const QVariantMap &params)
                      this, SLOT(onUrlChanged(const QUrl&)));
 
     /* set a per-identity cookie jar on the page */
-    uint identity = 0;
-    if (params.contains(SSOUI_KEY_IDENTITY)) {
-        identity = params.value(SSOUI_KEY_IDENTITY).toUInt();
-    }
+    uint identity = q->identity();
     CookieJarManager *cookieJarManager = CookieJarManager::instance();
     CookieJar *cookieJar = cookieJarManager->cookieJarForIdentity(identity);
     addBrowserCookies(cookieJar);
@@ -421,10 +479,14 @@ QWidget *BrowserRequestPrivate::buildWebViewPage(const QVariantMap &params)
         page->setAllowedSchemes(QStringList("https"));
     }
 
+    m_ignoreSslErrors = clientData.value(keyIgnoreSslErrors, false).toBool();
+
     QUrl url(params.value(SSOUI_KEY_OPENURL).toString());
     setupViewForUrl(url);
     QObject::connect(m_webView, SIGNAL(urlChanged(const QUrl&)),
                      this, SLOT(onUrlChanged(const QUrl&)));
+    QObject::connect(m_webView, SIGNAL(loadProgress(int)),
+                     this, SLOT(onLoadProgress()));
     QObject::connect(m_webView, SIGNAL(loadFinished(bool)),
                      this, SLOT(onLoadFinished(bool)));
     m_webViewLayout->addWidget(m_webView);
@@ -564,6 +626,18 @@ void BrowserRequestPrivate::showDialog()
     q->setWidget(m_dialog);
 }
 
+static Qt::ScrollBarPolicy scrollPolicyFromValue(const QVariant &value)
+{
+    QString stringValue = value.toString();
+    if (stringValue == valueAlwaysOn) {
+        return Qt::ScrollBarAlwaysOn;
+    } else if (stringValue == valueAlwaysOff) {
+        return Qt::ScrollBarAlwaysOff;
+    } else {
+        return Qt::ScrollBarAsNeeded;
+    }
+}
+
 void BrowserRequestPrivate::setupViewForUrl(const QUrl &url)
 {
     QString host = url.host();
@@ -602,10 +676,23 @@ void BrowserRequestPrivate::setupViewForUrl(const QUrl &url)
         m_webView->setZoomFactor(m_settings->value(keyZoomFactor).toReal());
     }
 
+    if (m_settings->contains(keyHorizontalScrollBar)) {
+        Qt::ScrollBarPolicy policy =
+            scrollPolicyFromValue(m_settings->value(keyHorizontalScrollBar));
+        page->mainFrame()->setScrollBarPolicy(Qt::Horizontal, policy);
+    }
+
+    if (m_settings->contains(keyVerticalScrollBar)) {
+        Qt::ScrollBarPolicy policy =
+            scrollPolicyFromValue(m_settings->value(keyVerticalScrollBar));
+        page->mainFrame()->setScrollBarPolicy(Qt::Vertical, policy);
+    }
+
     page->setExternalLinksPattern(m_settings->value(keyExternalLinksPattern).
                                   toString());
     page->setInternalLinksPattern(m_settings->value(keyInternalLinksPattern).
                                   toString());
+    page->setAllowedUrls(m_settings->value(keyAllowedUrls).toString());
 }
 
 void BrowserRequestPrivate::notifyAuthCompleted()
@@ -643,7 +730,9 @@ QWebElement BrowserRequestPrivate::initializeField(const QString &settingsKey,
         const QVariantMap &params = q->parameters();
         if (!paramKey.isEmpty() && params.contains(paramKey)) {
             QString value = params.value(paramKey).toString();
-            element.setAttribute("value", value);
+            if (!value.isEmpty()) {
+                element.setAttribute("value", value);
+            }
         }
     } else {
         BLAME() << "Couldn't find element:" << selector;
